/*
 * Copyright E. Bigeon (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * Socket implementation of GCLC.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
package fr.bigeon.gclc.socket;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.ConsoleApplication;
import fr.bigeon.gclc.manager.ConsoleManager;
import fr.bigeon.smu.StringEncoder;

/** This is a socket communicating console consoleManager
 * <p>
 * To use this application, the following flow should be used:
 *
 * <pre>
 * SocketConsoleApplicationShell shell = new SocketConsoleApplicationShell();
 * ConsoleApplication myApplication = new MyConsoleApplication(shell.getConsoleManager(), ...);
 * shell.setApplication(myApplication);
 * Thread th = new Thread(shell);
 * th.start();
 * </pre>
 * <p>
 * This will start the application in a separate thread. The application will be
 * listening on the given socket and writing back on it. If this is all your
 * application, you should then {@link Thread#join()} the thread to wait for the
 * end of the execution.
 *
 * @author Emmanuel Bigeon */
public class SocketConsoleApplicationShell implements Runnable {

    /**
     * 
     */
    private static final String INTERRUPTION_WHILE_WORKING = "Interruption while application was working"; //$NON-NLS-1$
    /** The end of line character */
    protected static final String EOL = "\n"; //$NON-NLS-1$
    /** The encoder */
    private static final StringEncoder ENCODER = new StringEncoder("%", //$NON-NLS-1$
            Arrays.asList(EOL));
    /** The class logger */
    private static final Logger LOGGER = Logger
            .getLogger(SocketConsoleApplicationShell.class.getName());
    /** The listening port */
    private final int port;
    /** The input */
    private final PipedInputStream consoleInput = new PipedInputStream();
    /** The application */
    private ConsoleApplication app;
    /** The session closing command */
    private final String close;
    /** The running status */
    private boolean running;
    /** An object to lock on for prompt */
    private final Object promptingLock = new Object();

    /** The console manager implementation */
    private final ThreadedServerConsoleManager consoleManager = new ThreadedServerConsoleManager(
            ENCODER, promptingLock);
    /** The auto close flag. if this is true, every request closes the session
     * after its call */
    private final boolean autoClose;
    /** The server socket */
    private ServerSocket serverSocket;
    /** The application shutdown string */
    private final String applicationShutdown;

    /** Create a socket application shell which will listen on the given port
     * and close session upon the provided string reception by client
     *
     * @param port the port to listen to
     * @param close the session closing command
     * @param applicationShutdown the appication shut down command */
    public SocketConsoleApplicationShell(int port, String close,
            String applicationShutdown) {
        this.port = port;
        this.close = close;
        this.applicationShutdown = applicationShutdown;
        this.autoClose = false;
    }

    /** Create a socket application shell which will listen on the given port
     * and auto close session after one instruction
     *
     * @param port the port to listen to
     * @param autoClose if the session must be closed once the request has been
     *            sent
     * @param applicationShutdown the application shutdown command */
    public SocketConsoleApplicationShell(int port, boolean autoClose,
            String applicationShutdown) {
        this.port = port;
        this.autoClose = autoClose;
        this.applicationShutdown = applicationShutdown;
        this.close = autoClose ? null : "close"; //$NON-NLS-1$
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run() */
    @Override
    public void run() {
        // Create the server
        try (ServerSocket actualServerSocket = new ServerSocket(port)) {
            this.serverSocket = actualServerSocket;
            running = true;
            // Create the streams
            try (PipedOutputStream outStream = new PipedOutputStream();
                 BufferedWriter writer = new BufferedWriter(
                         new OutputStreamWriter(outStream));
                 InputStreamReader isr = new InputStreamReader(consoleInput);
                 BufferedReader inBuf = new BufferedReader(isr)) {
                consoleInput.connect(outStream);
                consoleManager.setInput(inBuf);
                runSokectServer(writer);
                // Close the application
                // Pass command to application
                if (app.isRunning()) {
                    writer.write(applicationShutdown + EOL);
                    writer.flush();
                }
            }
        } catch (

        final IOException e) {
            LOGGER.log(Level.SEVERE,
                    "Communication error between client and server", e); //$NON-NLS-1$
        }
    }

    /** @param writer the writer to the application
     * @throws IOException if the communication with the client failed */
    private void runSokectServer(BufferedWriter writer) throws IOException {
        final ConsoleRunnable runnable = new ConsoleRunnable(app,
                promptingLock);
        Thread appThOld = null;
        Thread appThNext = new Thread(runnable);
        while (running) {
            try (Socket clientSocket = serverSocket.accept();
                 PrintWriter out = new PrintWriter(
                         clientSocket.getOutputStream(), true);
                 InputStreamReader isr = new InputStreamReader(
                         clientSocket.getInputStream());
                 BufferedReader in = new BufferedReader(isr);) {
                // this is not threaded to avoid several clients at the same
                // time
                consoleManager.setOutput(out);
                // Initiate application
                if (appThOld == null || !appThOld.isAlive()) {
                    appThNext.start();
                    // Prepare next start
                    appThOld = appThNext;
                    appThNext = new Thread(runnable);
                } else {
                    out.println("Reconnected"); //$NON-NLS-1$
                }
                communicate(writer, in);
            } catch (SocketException e) {
                LOGGER.log(Level.INFO, "Socket closed", e); //$NON-NLS-1$
            }
        }
    }

    /** active communication between server and client
     * 
     * @param writer the writer to the application
     * @param in the input from the client
     * @throws IOException if the communication failed */
    private void communicate(BufferedWriter writer,
                             BufferedReader in) throws IOException {
        synchronized (promptingLock) {
            if (!consoleManager.isPrompting()) {
                try {
                    // wait for application to finish its operation
                    promptingLock.wait();
                } catch (final InterruptedException e) {
                    LOGGER.log(Level.SEVERE, INTERRUPTION_WHILE_WORKING, e);
                }
            }
            if (autoClose) {
                communicateOnce(in, writer);
            } else {
                communicateLoop(in, writer);
            }
        }
    }

    /** @param in the input from the client
     * @param writer the output to the client
     * @throws IOException if the communication failed */
    private void communicateOnce(BufferedReader in,
                                 BufferedWriter writer) throws IOException {
        String ln;
        if ((ln = in.readLine()) != null) {
            if (ln.equals(close)) {
                return;
            }
            // Pass command to application
            writer.write(ln + EOL);
            writer.flush();
            try {
                // Wait for application process to
                // finish
                promptingLock.wait();
            } catch (final InterruptedException e) {
                LOGGER.log(Level.SEVERE, INTERRUPTION_WHILE_WORKING, e);
            }
        }
    }

    /** @param in the input from the client
     * @param writer the output to the client
     * @throws IOException if the communication failed */
    private void communicateLoop(BufferedReader in,
                                 BufferedWriter writer) throws IOException {
        String ln;
        while (app.isRunning() && (ln = in.readLine()) != null) {
            if (ln.equals(close)) {
                break;
            }
            // Pass command to application
            writer.write(ln + EOL);
            writer.flush();
            try {
                // Wait for application process to
                // finish
                promptingLock.wait();
            } catch (final InterruptedException e) {
                LOGGER.log(Level.SEVERE, INTERRUPTION_WHILE_WORKING, e);
            }
        }
    }

    /** @return the consoleManager */
    public synchronized ConsoleManager getConsoleManager() {
        return consoleManager;
    }

    /** @param app the application to set */
    public synchronized void setApplication(ConsoleApplication app) {
        this.app = app;
    }

    /** This method will request the server to stop.
     * <p>
     * In most cases, this will terminate communication on every client. On some
     * cases, the */
    public void stop() {
        running = false;
        try {
            serverSocket.close();
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Exception in closing socket server", e); //$NON-NLS-1$
        }
        synchronized (promptingLock) {
            promptingLock.notifyAll();
        }
        app.exit();
    }

}
